=== modified file 'cgroup-reap-all.c'
--- cgroup-reap-all.c	2014-08-26 19:52:27 +0000
+++ cgroup-reap-all.c	2014-11-20 20:33:59 +0000
@@ -59,7 +59,7 @@
 		g_list_free(pidlist);
 	}
 
-	g_clear_object(&cgmanager);
+	cgroup_manager_unref(cgmanager);
 
 	return 0;
 }

=== modified file 'helpers-shared.c'
--- helpers-shared.c	2014-11-20 17:26:18 +0000
+++ helpers-shared.c	2014-11-20 20:33:59 +0000
@@ -171,6 +171,7 @@
 GDBusConnection *
 cgroup_manager_connection (void)
 {
+	gboolean use_session_bus = g_getenv("UBUNTU_APP_LAUNCH_CG_MANAGER_SESSION_BUS") != NULL;
 	GMainContext * context = g_main_context_new();
 	g_main_context_push_thread_default(context);
 
@@ -185,7 +186,7 @@
 	guint timeout = g_source_attach(timesrc, context);
 	g_source_unref(timesrc);
 
-	if (g_getenv("UBUNTU_APP_LAUNCH_CG_MANAGER_SESSION_BUS")) {
+	if (use_session_bus) {
 		/* For working dbusmock */
 		g_debug("Connecting to CG Manager on session bus");
 		g_bus_get(G_BUS_TYPE_SESSION,
@@ -209,11 +210,45 @@
 	g_object_unref(connection.cancel);
 
 	g_main_context_pop_thread_default(context);
-	g_main_context_unref(context);
+
+	if (!use_session_bus && connection.con != NULL) {
+		g_object_set_data(G_OBJECT(connection.con), "cgmanager-context", context);
+	} else {
+		g_main_context_unref(context);
+	}
 
 	return connection.con;
 }
 
+static void
+cgroup_manager_unref_weak (gpointer mainp, GObject * old_obj)
+{
+	g_main_loop_quit((GMainLoop *)mainp);
+}
+
+void
+cgroup_manager_unref (GDBusConnection * cgmanager)
+{
+	GMainContext * creationcontext = g_object_get_data(G_OBJECT(cgmanager), "cgmanager-context");
+	if (creationcontext == NULL) {
+		g_object_unref(cgmanager);
+		return;
+	}
+
+	g_main_context_push_thread_default(creationcontext);
+
+	GMainLoop * shutdownloop = g_main_loop_new(creationcontext, FALSE);
+	g_object_weak_ref(G_OBJECT(cgmanager), cgroup_manager_unref_weak, shutdownloop);
+
+	g_object_unref(cgmanager);
+	g_main_loop_run(shutdownloop);
+
+	g_main_loop_unref(shutdownloop);
+
+	g_main_context_pop_thread_default(creationcontext);
+	g_main_context_unref(creationcontext);
+}
+
 /* Get the PIDs for a particular cgroup */
 /* We're using the base cgroup 'freezer' in this code (and
    in the Upstart jobs). Really the actual group is meaningless

=== modified file 'helpers.h'
--- helpers.h	2014-08-08 19:50:53 +0000
+++ helpers.h	2014-11-20 20:33:59 +0000
@@ -49,6 +49,7 @@
 void      starting_handshake_wait        (handshake_t *   handshake);
 
 GDBusConnection * cgroup_manager_connection (void);
+void              cgroup_manager_unref (GDBusConnection * cgroup_manager);
 GList *   pids_from_cgroup       (GDBusConnection * cgmanager,
                                   const gchar *   jobname,
                                   const gchar *   instancename);

=== modified file 'libubuntu-app-launch/ubuntu-app-launch.c'
--- libubuntu-app-launch/ubuntu-app-launch.c	2014-09-30 02:16:15 +0000
+++ libubuntu-app-launch/ubuntu-app-launch.c	2014-11-20 20:33:59 +0000
@@ -1301,7 +1301,7 @@
 
 	if (is_click(appid)) {
 		GList * pids = pids_from_cgroup(cgmanager, "application-click", appid);
-		g_clear_object(&cgmanager);
+		cgroup_manager_unref(cgmanager);
 
 		ual_tracepoint(pids_list_finished, appid, g_list_length(pids));
 		return pids;
@@ -1309,7 +1309,7 @@
 		gchar * jobname = g_strdup_printf("%s-", appid);
 		GList * pids = pids_from_cgroup(cgmanager, "application-legacy", jobname);
 		g_free(jobname);
-		g_clear_object(&cgmanager);
+		cgroup_manager_unref(cgmanager);
 
 		ual_tracepoint(pids_list_finished, appid, g_list_length(pids));
 		return pids;
@@ -1340,7 +1340,7 @@
 	g_array_free(apps, TRUE);
 	g_object_unref(con);
 
-	g_clear_object(&cgmanager);
+	cgroup_manager_unref(cgmanager);
 
 	ual_tracepoint(pids_list_finished, appid, g_list_length(pids));
 	return pids;

